/*
**  VFDシールド　プログラム
**
**  ノリタケ伊勢電子製　GU-7903シリーズを同期式シリアル接続で
**  つないでみました。
**
**   UnicodeをシフトJISに変換するサンプルを追加しました。　
**   対応コードは X0208 です。
**   
**   GU-7903モジュールのジャンパー設定(J2ショート）が必要です。
**
*/

#include <avr/pgmspace.h>    // PROGMEM用ライブラリ読み込み
#include "table.h"            // コード変換用テーブル読み込み

//   ピン配置

int vfd_reset =  8;  //  VFDコネクタ 6番ピン
int vfd_si    =  9;  //　VFDコネクタ 2番ピン
int vfd_sbusy = 10;  //　VFDコネクタ 4番ピン
int vfd_clk   = 11;  //　VFDコネクタ 5番ピン
int sw_blue   = 12;  //  青ボタン
int sw_red    = 13;  //　赤ボタン

// 作成した関数のプロトタイプ

void vfd_print(unsigned char message[],int L);      //　Lバイトのバイト列を表示器に送信します。
void vfd_put(unsigned char unebyte);                //　１バイトを表示器に送信します。
void vfd_kanjiOn();                                 //　JIS漢字モードを選択します。
void vfd_crlf();                                    //　改行コードを送信します。
void vfd_setCursor(int X, int Y);                   //　カーソル位置（書き込み位置）を移動します。
unsigned int  getSJIS( unsigned int Unicode);       //　UTF-16コードのUnicode文字をShift-JISに変換します。
void vfd_print_hex(unsigned char message[],int L);  //　Lバイトのバイト列を16進で表示器に送信します。
void vfd_print_utf8(char message[]);                //　UTF8エンコードされたUnicodeテキストをS-JISに変換して表示機に送信します。
void vfd_put_SJIS(unsigned int charact);            //　S-JISエンコードされた文字を表示器に送ります。
void vfd_put_hex1( unsigned char ch);               //　1バイトの文字を16進表記で表示機に送ります。
void vfd_VScroll();                                 //　縦スクロールモードを選択します。
void vfd_HScroll(char scrollSpeed);                 //　横スクロールモードを選択し、スクロール速度を設定します。（数字が大きくなると遅くなります。）

// the setup routine runs once when you press reset:
// リセット後に１回だけ実行されるプログラム
void setup() {                

      int key;
      char  kanji_code[] =  "蛍光表示管";
    
      // initialize the digital pins
      pinMode(vfd_reset, OUTPUT);
      pinMode(vfd_clk, OUTPUT);
      pinMode(vfd_si, OUTPUT);
      pinMode(vfd_sbusy, INPUT);
      pinMode(sw_blue, INPUT_PULLUP);
      pinMode(sw_red, INPUT_PULLUP);
      digitalWrite(vfd_reset,LOW);
      delay(1);
      digitalWrite(vfd_reset,HIGH);
      
      vfd_VScroll();
      unsigned char message[12] = "Hello World";
      vfd_print(message, sizeof(message)-1);
    
      vfd_crlf();                   //　改行  
      vfd_kanjiOn();                //　漢字表示モード設定
      vfd_HScroll(1);               //　横スクロール指定（スピード=１）
        
      vfd_print_utf8("                比較的安価でC言語風の統合開発環境が整ったマイコンボードArduino用にVFDシールドを作成して表示させてみました。");
      vfd_print_utf8("使用したのは、AVRマイクロコントローラを搭載したマイコンボード、Arduino UNO（アルドゥイーノ ウノ）。 PCとUSB接続し、専用のArduino IDE（開発環境）を使えば比較的に容易にプログラムできます。");
      vfd_print_utf8("Arduinoのハードウェア的な機能を拡張させたりするドーターボード（アドオンボード）のことをシールド(Shield)と呼び、いろいろなものが作成されたり、販売されたりしています。 ");    // これでもOK
      vfd_print_utf8("Arduinoに各種シールドを繋ぐ（実際はコネクタを使って合体させる）と、例えばDCモーターやステッピングモーターを制御できるようになったりと、Arduinoの活用幅が広がります。");    // これでもOK

      delay(2000);                 // 2秒停止

      vfd_put(0x0c);               // 画面クリア
      vfd_setCursor(0,2);          //カーソル位置（文字書き込み位置）移動
      //    "蛍光表示管"  
      vfd_print_utf8(kanji_code);  // 配列に定義した文字列です。
}

// the loop routine runs over and over again forever:
// 繰り返し実行されるプログラム
void loop() {

    if(digitalRead(sw_red) == LOW) //赤ボタン
       { vfd_setCursor(100,1); // カーソル移動
        vfd_put(0x90);       //　”赤”
        vfd_put(0xd4);
        vfd_put((0x31));
      }
    if(digitalRead(sw_blue) == LOW) //青ボタン
      {  vfd_setCursor(100,1); // カーソル移動
        vfd_put(0x90);        //　”青”
        vfd_put(0xc2);
        vfd_put(getSJIS(0x30));
      }
}

void vfd_HScroll(char scrollSpeed)
{
   unsigned char data[5];
    data[0] = 0x1f; 
    data[1] = 0x03;
    data[2] = 0x1f;
    data[3] = 0x73;
    data[4] = scrollSpeed;
  
    vfd_print(data,6);  
}

void vfd_print(unsigned char message[], int L)
{
 int i;
 for(i=0; i<L; i++)
  {
     vfd_put(message[i]);
  } 
}
void vfd_print_hex(unsigned char message[],int L)
{
 int i;
 int aa;
 
 for(i=0; i<L; i++)
  {
    if((aa =  message[i] / 16) > 9) {aa += 7;}
     vfd_put( aa + '0');
    if((aa =  message[i] % 16) > 9) {aa += 7;}
     vfd_put( aa + '0');
     vfd_put(' ');
  } 
}
void vfd_print_utf8(char message[])
{
 int i=0;
 unsigned int aa;
  do
 {
   if( (message[i] & 0x80) == 0)  // 1 byte
   {
      vfd_put_SJIS(getSJIS(message[i]));
     i++;
   }
   else
   {
     if((message[i] & 0xe0) == 0xc0)  // 2 byte
     {
    
     aa = ((unsigned int)message[i] & 0x1f) * 0x40;
     aa += message[i+1] & 0x3f;
     
     vfd_put_SJIS(getSJIS(aa));
     
       i += 2;
     }
     else
     {
        if( (message[i] & 0xf0) == 0xe0)  // 3 byte
        {
              aa = (unsigned int) message[i];
              aa *=  0x1000 ;
              aa += ((unsigned int)message[i+1] & 0x3f) * 0x40;
              aa  += (message[i+2] & 0x3f);
              aa = getSJIS(aa);
              vfd_put_SJIS(aa);
              
            i += 3; 
        }
        if( (message[i] & 0xf8) == 0xf0)  // 4 byte
        {
          i += 4;
        }
        else
        {
          if( (message[i] & 0xfc) == 0xf8)  // 5 byte
          {
            i += 5;
          }
          else
          if( (message[i] & 0xfe) == 0xfc)  // 6 byte
            i+= 6;
        }
      }
   }
 } while (message[i] != 0) ;
}

void vfd_put_SJIS(unsigned int charact)
{
  if(charact > 255) { vfd_put(charact / 256);}
  vfd_put(charact % 256);
}

void vfd_put_hex1( unsigned char ch)
{
 unsigned int aa;

  if (( aa = ch / 16 ) >=10) aa += 7;
  vfd_put(aa + '0');
  if (( aa = ch % 16 ) >=10) aa += 7;
  vfd_put(aa + '0');
  vfd_put(' ');
}

void vfd_put(unsigned char onebyte)
{
    while( digitalRead(vfd_sbusy) == HIGH){}
    shiftOut(vfd_si, vfd_clk, LSBFIRST, onebyte);  
}
void vfd_crlf()
{
  shiftOut(vfd_si, vfd_clk, LSBFIRST, 0x0d);
  shiftOut(vfd_si, vfd_clk, LSBFIRST, 0x0a);
}

void vfd_VScroll()
{
  vfd_put(0x1f);
  vfd_put(0x02);
}
void vfd_kanjiOn()
{
 unsigned char data[15];
   data[0]=0x1f;
   data[1]=0x28;
   data[2]=0x67;
   data[3]=0x01;
   data[4]=0x02;
   data[5]=0x1f;
   data[6]=0x28;
   data[7]=0x67;
   data[8]=0x02;
   data[9]=0x01;
   data[10]=0x1f;
   data[11]=0x28;
   data[12]=0x67;
   data[13]=0x0f;
   data[14]=0x00;
 
  vfd_print(data,15);
}

void vfd_setCursor(int X, int Y)
{
   unsigned char data[6];
    data[0] = 0x1f; 
    data[1] = 0x24;
    data[2] = X % 256;
    data[3] = X / 256;
    data[4] = Y % 256;
    data[5] = Y / 256;
  
    vfd_print(data,6); 
}

unsigned int getSJIS( unsigned int Unicode)
{
            unsigned int i, j;
            unsigned int uucode;
            unsigned int key;
                  
            key = 0;
            uucode = Unicode;
            if (Unicode < 162) { return Unicode; } // ASCII はそのまま返す
            // look up 1st table
            for (i = 0; i < 114; i++)
            {
              if (uucode < uniTable[i])
                {
                    // found

                    if (tableSize[i] == 0)
                    {  // 計算テーブル

                        switch (i)
                        {
                            case 2:
                            case 4:
                                return Unicode + 32782;
                                break;
                            case 3:
                            case 5:
                                return Unicode + 32781;
                                break;
                            case 7:
                                return Unicode + 32816;
                                break;
                            case 8:
                                return Unicode + 32817;
                                break;
                            case 9:
                                return Unicode + 32832;
                                break;

                            case 10:
                                return Unicode + 32833;
                                break;
                            case 11:
                                return Unicode + 32834;
                                break;
                            case 19:
                                return Unicode + 20841;
                                break;
                            case 21:
                                return Unicode + 21086;
                                break;
                            case 23:
                                return Unicode + 21151;
                                break;
                            case 24:
                                return Unicode + 21152;
                                break;
                            case 107:
                            case 109:
                                return Unicode - 31937;
                                break;
                            case 111:
                                return Unicode - 31936;
                                break;
                            case 113:
                                return Unicode - 65216;
                                break;
                            default:
                                return 0; // Error
                        }
                    }
                    else
                    {  // サーチテーブル
                        for (j = 0; j < tableSize[i]; j++)
                        {
                          if (uucode == pgm_read_byte_near( table +key++))
                            {
                                key --;
                                return (pgm_read_byte_near(table + (key * 2) + 6525 )) * 256 + pgm_read_byte_near(table + (key * 2) + 6526);  // found
                            }
                        }
                        return 0;  // not on the table
                    }
                }
                uucode -= uniTable[i];
                key += tableSize[i];
            }
            return 0;  // Not found
        }

